# Gcode_importer.rb - Imports gcode .gcode file in SketchUp
#
# Copyright (C) 2013 David Fishelzon
#
# License: Apache License, Version 2.0
#
# G-Code Importer 
#
require 'sketchup'

module GcodeImporterModule

  class GcodeImporter < Sketchup::Importer
    FilamentWidth = 3.mm
    NozelDiameter = 0.5.mm
    FilaCutArea   = Math::PI * ( FilamentWidth*FilamentWidth / 4 ) 

    # Find the "filament" component if it exist OR
    # if the make a "filament" component if it does not exist
    def Get_Filament_component
      @model        = Sketchup.active_model
      @filamentComp = @model.definitions[ "filament" ]
      unless @filamentComp
        # if there is no existing component named "filament", make one
        # We will make one
        layerHight    = 1.0.mm
        @filamentComp = @model.definitions.add("filament")
        @pts          = [  ORIGIN,
                 [-NozelDiameter/2.0, 0.0.mm , -layerHight/2.0], 
                 [-NozelDiameter/2.0, 0.0.mm , -layerHight    ], 
                 [ NozelDiameter/2.0, 0.0.mm , -layerHight    ], 
                 [ NozelDiameter/2.0, 0.0.mm , -layerHight/2.0] ]
        @newface = @filamentComp.entities.add_face(@pts)
        @newface.casts_shadows    = false
        @newface.receives_shadows = false
        @newface.reverse! if @newface.normal.y < 0
        @newface.pushpull(1.0.mm)
        @filamentComp.entities.each { |@ent|
          if @ent.class == Sketchup::Edge
             @ent.hidden        = true
             @ent.casts_shadows = false
          end # if edge
        } # each ent
      end # unless @filamentComp
      @scaleZconst     = FilaCutArea / NozelDiameter / @filamentComp.bounds.depth

      @HeadMoveComp   = @model.definitions[    "HeadMove" ]
      unless @HeadMoveComp
        @HeadMoveComp = @model.definitions.add("HeadMove")
        @oldPoint     = [0.0, 0.0   , 0.0]
        @newPoint     = [0.0, 1.0.mm, 0.0]
        @HeadMoveComp.entities.add_line @oldPoint, @newPoint
        @HeadMoveComp.entities[0].casts_shadows = false
      end # unless @HeadMoveComp
    end # def Get_Filament_component

    # Adds a "filament" component between @oldPoint and @newPoint
    def moveOld2newPoint( comp )
      @lineVector  = @oldPoint.vector_to @newPoint
      return nil if @lineVector.length.to_f < 1e-6 

      # Scaling and adding an instance of the filament
      @scaleL      = @lineVector.length.to_f / @filamentComp.bounds.height   #width #depth 
      return nil if @scaleL < 1e-6 or @scaleZ < 1e-6
      @trans       = Geom::Transformation.scaling 1.0, @scaleL, @scaleZ
      @instance    = @model.active_entities.add_instance( comp, @trans )

      # if needed rotate the filament
      @lineVector.normalize!
      @instYaxis   = Geom::Vector3d.new( 0.0, @filamentComp.bounds.height, 0.0 ).normalize #height givs the Y axis
      @rotDeg      = @instYaxis.dot( @lineVector )
      @rotDeg      = Math::acos( @rotDeg )
      if @rotDeg.to_f - 1e-6 > 0.0
        if @rotDeg.to_f + 1e-6 > Math::PI 
          @rotVector = Geom::Vector3d.new( 0.0, 0.0, 1.0 )
        else # if PI
          @rotVector = @instYaxis.cross( @lineVector )
        end # if PI
        @trans       = Geom::Transformation.rotation( ORIGIN, @rotVector, @rotDeg )
        @instance.transform!( @trans )
      end # if @rotDeg > 0.0

      # Moving the comp to the @oldPoint
      if not ( ORIGIN == @oldPoint )
        @trans       = Geom::Transformation.new @oldPoint 
        @instance.transform!( @trans )
      end #if not ORIGEN

      @instance.casts_shadows    = false
      @instance.receives_shadows = false
      return @instance
    end # def move_old2new

    def initialize
      Get_Filament_component()      # Add a filament component
    end

    def description
      "Gcode Importer (*.gcode;*.bfb) by David Fishelzon"
    end

    def id
      "DavidFishelzon@Yahoo.com/Gcode_importer"
    end

    def file_extension
      return "*.gcode;*.bfb"
    end

    def supports_options?
      false
    end

    def do_options
      false
    end

    def load_file( filename, status )
      if status == true
        begin
          if (Sketchup.version.to_i < 7)
            @model.start_operation("Import G-code")
          else
            @model.start_operation("Import G-code",true)
          end
          @model.active_view.dynamic = true
          @model.description = " This is an import of the G-Code file :" + filename.scan(/[[:print:]]/).join
          gcode_import(filename)
          @model.commit_operation # commits an operation for undo
          @model.active_view.zoom_extents
          @model.active_view.refresh     
          puts "FINISH FINISH FINISH FINISH"
          Sketchup.status_text=""
          return 0 # 0 = Import Success (no additional dialog shown)
        rescue Exception=>e
          puts "!!!!!!! !!!!!!!!!!!!!!!!!!!!!!!!! !!!!!!!"
          puts "!!!!!!! Bug in the Gcode importer !!!!!!!"
          puts "!!!!!!! !!!!!!!!!!!!!!!!!!!!!!!!! !!!!!!!"
          puts e.message
          puts e.backtrace
          UI.messagebox "Bug in the Gcode importer"
          return 1 #1 = Import Failed (no additional dialog shown)
        end # begin rescue
      end # if status
    end # def load_file

    def gcode_import( filename )
      @model       = Sketchup.active_model
      @currentLayer= @model.active_layer
      @x = @y = @z = @f = 0.0       # x,y,z are positions f is the speed of movement
      @units       = 1.0.mm         # Default units is mm
      @e           = 1.0.mm # extruder on by default 
      @oldE        = @e     # in absolute length we need to get only the extruded length
      @incremental = false
      @extruderMode= "Absolute"
      @oldPoint    = @newPoint = [ @x, @y, @z ]
      @lineNumber  = 0
      @layerUsed   = false
      @scaleZ      = 0.1 / @filamentComp.bounds.depth 
      @model.active_layer = @model.layers.add "Gcode Line #{"%8d" % @lineNumber}"
      Sketchup.status_text=" Starting G Code Import "

      IO.foreach(filename) do |@line|
        @lineNumber += 1
        if @lineNumber % 1000 == 0
          Sketchup.status_text=" #{@lineNumber} Lines imported "
          if @lineNumber == 10000
            @model.active_view.zoom_extents  
            @model.active_view.refresh
          end # if @lineNumber == 10000 
        end #if
 
        @line = @line.scan(/[[:print:]]/).join # remuve all non printable chars from the line
        @line = @line[/[^;]*/].strip           # remuve the ";" remarks and the start & end spaces 
        next if @line.length < 2               # not enuf chars in this line move to next line

        # execute the Command:
        @command, *@params = @line.split
        case @command

        when 'G1', 'G28', 'G92', 'M92' # G1 move motors: 
          # read the line by adding, befor the first number(/\d/), an '=' and evaluating 
          @params.map {|@var| 
            eval('@' + @var.insert(@var.index(/[-\d]/),'=').downcase + '*@units' ) if @var[/[xyzefXYZEF]/] 
          } # @params.map {|@var|

          if @incremental == true
            # convert to absolute positioning
            @newPoint = [@x+@oldPoint[0],@y+@oldPoint[1],@z+@oldPoint[2]]
          else 
            # absolute positioning
            @newPoint = [@x             ,@y             ,@z             ]
          end  # if @incremental 

          # New Layer needed?
          if not (@z - @oldPoint[2]).between?( -1e-6, 1e-6)
            # movement in the Z direction 
            if @layerUsed == true
              if @extruderMode == "ON"
                @scaleZ           = ( @z - @oldPoint[2] ) / @filamentComp.bounds.depth 
              end # if "ON"
              @model.active_layer = @model.layers.add "Gcode Line #{"%8d" % @lineNumber}"
              @layerUsed          = false
            end # if @layerUsed
          end  # if 

          # fix @e extruded length
          if @extruderMode == "Absolute" 
            # new absolute e - extruded length
            # G92 sets an origin offset - Don't use @oldE
            @oldE = @e if @command == 'G92' 

            # if we do not have new e in this line use the old one
            @e    = @oldE if not @line[/[eE]/]

            # calculate relative extruded length
            @e    = @e - @oldE 
            @oldE = @e + @oldE
          end # if @extruderMode == "Absolute" 

          @e = 0.0 if @e < 0.0 # don't pull back the filament

          # add the correct component
          if @e <= 0.00001.mm.to_f or @extruderMode == "OFF"
            # No extruded filament - so just draw a line
            @lastMove    = moveOld2newPoint( @HeadMoveComp )
          else # @e == 0.0
            # Extruded filament
            if not @extruderMode == "ON"
              # @extruderMode == "Absolute" or "Incremental" 
              # calc the @scaleZ from E - the extruded length
              moveLength   = Geom::Vector3d.new( 
                                   @newPoint[0] - @oldPoint[0], 
                                   @newPoint[1] - @oldPoint[1], 
                                           0.0 ).length.to_f
              if moveLength > 0.001.mm.to_f
                # move in X,Y - fix the @scaleZ according to @e
                @scaleZ    = @e / moveLength * @scaleZconst
              end #if moveLength > 0.0
            end  #if @extruderMode == "ON"

            # don't move in the Z axis
            saveZ        = @newPoint[2]
            @newPoint[2] = @oldPoint[2]
            @lastMove    = moveOld2newPoint( @filamentComp )
            @newPoint[2] = saveZ

            # Mark this Layer as Used if we put a filament component in it
            @layerUsed   = true if @lastMove
          end  # @e == 1e-6

          # if last head move was succesful
          if @lastMove
            @oldPoint     = @newPoint # save old location
#            @lastMove.name= "Line #{@lineNumber}: #{@line} @oldE>#{"%0.5f" % (@oldE/1.mm)}< e>#{"%0.5f" % (@e/1.mm)}<#{@extruderMode == "Absolute"?"Abs":"Rel"} "
            @lastMove.name= "Line #{@lineNumber}: #{@line} "
            @lastMove.set_attribute "physical_properties", "Move_Speed", @f/@units
          else # if lastMove
            # for some reason we could not add a component
            if not Geom::Vector3d.new(@newPoint[0] - @oldPoint[0],
                                      @newPoint[1] - @oldPoint[1],
                                      @newPoint[2] - @oldPoint[2]
                                     ).length.to_f.between?( -1e-6, 1e-6 )
              # it is not that the length was too short
              puts "#{@lineNumber} Unable to add line: >#{@line}< e=#{@e} "
              puts "x=#{@oldPoint[0]/@units} y=#{@oldPoint[1]/@units} z=#{@oldPoint[2]/@units} "
            end # if @line[/[xyzXYZ]/]
          end  # if lastMove
        ############################# end of when 'G1', 'G92', 'G28', 'M92' 
 
        when 'M103' # extruder off
          @extruderMode = "OFF"

        when 'M101' # extruder on 
          @extruderMode = "ON"

        when 'M82' # set extruder to absolute mode 
          @extruderMode = "Absolute"

        when 'M83' # set extruder to relative mode 
          @extruderMode = "Incremental"

        when 'G90' # G90 Absolute Positioning Implemented.
          @incremental = false

        when 'G91' # Incremental mode
          @incremental = true

        when 'G20'
          @units = 1.0.inch

        when 'G21'
          @units = 1.0.mm

        when 'M106', 'M107'
          #M106 Turn fan on.
          #M107 Turn fan off.
        when 'M104'
          #M104 S247.0 Set temperature to 145.0 DegC
        when 'M84'
          #Stop the idle hold on all axis and extruder - Noise reduction
#        when 'M108'
          # S400 Set Extruder 1 speed to S value/10 = 40rpm.
        when 'M109'
          #: Set Extruder Temperature and Wait

        else # of case command
          # disply a line what we did not understand
          puts "#{@lineNumber} New type of command: >#{@line}< "

        end # of case command
      end # of do |line|
      Sketchup.status_text="Done..."
      @model.active_layer = @currentLayer
    end # of def gcode_import( filename )

  end # class GcodeImporter < Sketchup::Importer

  Sketchup.register_importer(GcodeImporter.new)
end # GcodeImporterModule


